/*
 * OnTheFlyTester.java
 * Copyright (C) 2007,2008 Lars Frantzen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */


/*
 * This is the first and most simple scenario:
 *
 *                          Testing a single, passive Web Service
 *
 * Here, the SUT will never produce asynchronous outputs, only synchronous responses to calls.
 * Hence, it is sufficient to continously give synchronous or asynchronous inputs to the SUT, and analyse the
 * synchronous responses. No real quiescence here, and no unstable states (since a message cannot both be
 * synchronous and asynchronous.
 *
 */

package jambition;

import java.awt.Component;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;
import java.util.logging.Logger;
import javax.swing.JOptionPane;
import javax.swing.SwingWorker;
import stsimulator.STSimulator;
import stsimulator.STSimulator.EmptyInstantiatedLocationsException;
import stsimulator.interpreter.javacc.DumontParser;
import stsimulator.interpreter.javacc.DumontParser.RuntimeViolation;
import stsimulator.sts.InstantiatedMessage;
import stsimulator.sts.InstantiatedSwitch;
import stsimulator.sts.OperationKind;

/**
 * 
 * @author Lars Frantzen
 * @version Plastic
 */
public class OnTheFlyTester extends SwingWorker<Boolean, Void> {
    
    private TestEventDispatcher ted;
    private STSimulator sim;
    private Logger logger;
    private Random random = new Random();
    private Component parent;
    
    // flag: end the test loop because a final state is reached or a fault is discovered
    private boolean endTesting = false;
    
    private double oldScov = 0;
    private double oldLcov = 0;
    
    protected Boolean doInBackground() {
        try {
            startTestLoop();
        } catch (STSimulator.EmptyInstantiatedLocationsException ex) {
            logger.severe("------ FAILURE DETECTED ------");
            firePropertyChange("failure", false, true);
            JOptionPane.showMessageDialog(parent, "Failure Detected!", "Failure", JOptionPane.ERROR_MESSAGE);
        } catch (DumontParser.RuntimeViolation ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Simulator discovered a runtime violation!", "Error", JOptionPane.ERROR_MESSAGE);
        } catch (IOException ex) {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(parent, "Simulator had an IOException!", "Error", JOptionPane.ERROR_MESSAGE);
        }
        
        return true;
    }
    
    
    /** Creates a new instance of OnTheFlyTester */
    public OnTheFlyTester(Component parent, STSimulator sim, TestEventDispatcher ted, Logger logger) {
        this.parent = parent;
        this.ted = ted;
        this.sim = sim;
        this.logger = logger;
        if (Boolean.valueOf(Jambition.props.getProperty("DEBUGMODE")))
            logger.info(sim.getStatus());
    }
    
    private void startTestLoop() throws EmptyInstantiatedLocationsException, RuntimeViolation, IOException {
        while (!endTesting) {
            // compute next input
            InstantiatedMessage pendingInput = computeNextInput();
            // no further input specified
            if (pendingInput == null) {
                logger.info("No more inpus possible. SUT passed this test run.");
                endTesting = true;
            } else {
                logger.info("--> " + pendingInput.toString());
                // apply the computed input to the SSM
                giveInstantiatedMessageToSSM(pendingInput);
                // apply the computed input to the SUT
                if (pendingInput.getMessage().getOperation().getKind() == OperationKind.REQUESTRESPONSE) {
                    InstantiatedMessage pendingOutput = ted.invokeSync(pendingInput);
                    logger.info("<-- " + pendingOutput.toString());
                    giveInstantiatedMessageToSSM(pendingOutput);
                } else
                    ted.invokeOneWay(pendingInput);
            }
        }
    }
    
    public InstantiatedMessage computeNextInput() throws RuntimeViolation, IOException {
        // get all current input switches
        ArrayList<InstantiatedSwitch> inputs = new ArrayList<InstantiatedSwitch>(sim.getCurrentInputSwitches());
        
        boolean noSolutionFound = true;
        InstantiatedMessage nextInput = null;
        
        while (!inputs.isEmpty() && noSolutionFound) {
            InstantiatedSwitch candidate = (InstantiatedSwitch) inputs.get(random.nextInt(inputs.size()));
            // try to find a solution, if yes, fine, if not, remove the candidate
            nextInput = sim.findSolution(candidate);
            if (nextInput == null)
                inputs.remove(candidate);
            else
                noSolutionFound = false;
        }
        
        if (nextInput != null)
            return nextInput;
        else
            return null;
    }
    
    private void giveInstantiatedMessageToSSM(InstantiatedMessage im) throws EmptyInstantiatedLocationsException, RuntimeViolation {
        // addToLog("giving this testevent to the STS: " + te);
        sim.processInstantiatedMessage(im);
        if (Boolean.valueOf(Jambition.props.getProperty("DEBUGMODE")))
            logger.info(sim.getStatus());
        double newLcov = sim.getLocationCoverage();
        double newScov = sim.getSwitchCoverage();
        firePropertyChange("lcov", oldLcov, newLcov);
        firePropertyChange("scov", oldScov, newScov);
        oldLcov = newLcov;
        oldScov = newScov;
    }
    
    void stopTesting() {
        endTesting = true;
    }
}
